/*
								+--------------------------------+
								|                                |
								|  ***   Simple threading   ***  |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|            Thread.h            |
								|                                |
								+--------------------------------+
*/

/*
 *	2008-08-08
 *
 *	added #ifdef for windows 64
 *
 *	2008-09-17
 *
 *	fixed typo in word runable (there were two n's)
 *
 */

#ifndef __LAME_THREADS_INCLUDED
#define __LAME_THREADS_INCLUDED

/*
 *	class CRunable
 *		- virtual class for runable object
 */
class CRunable {
public:
	virtual void Run() = 0;
};

#if defined(WIN32) || defined (_WIN64)
#include <windows.h>
#else
#include <pthread.h>
#endif

/*
 *	class CThread
 *		- simple thread class; todo - create linux version
 */
class CThread {
protected:
	CRunable *m_p_runable;

#if defined(WIN32) || defined (_WIN64)
	HANDLE m_h_thread;
#else
	pthread_t m_t_thread; // thread
	pthread_mutex_t t_running_mutex;
	bool m_b_running; // thread state
#endif

	class CEmptyRunable : public CRunable {
		virtual void Run() {}
	};
	static CEmptyRunable empty_runable;

public:
	/*
	 *	CThread::CThread()
	 *		- default constructor, attaches empty runable object to a thread
	 */
	CThread();

	/*
	 *	CThread::CThread(CRunable &r_runable)
	 *		- constructor, attaches runable object to a thread
	 */
	CThread(CRunable &r_runable);

	/*
	 *	CThread::~CThread()
	 *		- destructor
	 */
	~CThread();

	/*
	 *	bool CThread::AttachRunable(CRunable &r_runable)
	 *		- attaches different runable, thread must not be running
	 *		- returns true on success, false on failure
	 */
	bool AttachRunable(CRunable &r_runable);

	/*
	 *	CRunable &CThread::r_Runable()
	 *		- returns reference to runable object
	 */
	CRunable &r_Runable();

	/*
	 *	const CRunable &CThread::r_Runable() const
	 *		- returns const reference to runable object
	 */
	const CRunable &r_Runable() const;

	/*
	 *	bool CThread::Start()
	 *		- starts the thread and returns immediately
	 *		- note this fails in case the thread is already running
	 *		- returns true on success, false on failure
	 */
	bool Start();

	/*
	 *	bool CThread::Run()
	 *		- starts the thread and waits untill it finishes
	 *		- note this fails in case the thread is already running
	 *		- also note this is not thread-safe (the Thread object doesn't change
	 *		  it's state to running while in this function and also can't be stopped)
	 *		- returns true on success, false on failure
	 */
	bool Run();

	/*
	 *	bool CThread::b_IsRunning()
	 *		- returns true if thread is running, otherwise returns false
	 */
	bool b_IsRunning();

	/*
	 *	bool CThread::Stop(bool b_force_kill = false)
	 *		- waits for the thread to end, if b_force_kill is set, it's ended forcibly
	 *		- returns true if thread is stopped, returns false if thread didn't
	 *		  stop in given time or on failure
	 */
	bool Stop(bool b_force_kill = false);

	/*
	 *	static int CThread::n_CPU_Num();
	 *		- utility function; returns number of CPU's
	 *		- in case CPU count cannot be determined, returns -1
	 */
	static int n_CPU_Num();

protected:
#if defined(WIN32) || defined (_WIN64)
	static unsigned long __stdcall _run(void *p_arg); // run for windows
#else
	static void *_run(void *p_arg); // run for linux
#endif
	const CThread &operator =(const CThread &r_thread) { return r_thread; } // can't copy threads this way
};

#endif //__LAME_THREADS_INCLUDED
